from pyopencl_extension.modifications_pyopencl import cltypes
import numpy as np
char2=cltypes.char2
char4=cltypes.char4
char8=cltypes.char8
char16=cltypes.char16
short2=cltypes.short2
short4=cltypes.short4
short8=cltypes.short8
short16=cltypes.short16
int2=cltypes.int2
int4=cltypes.int4
int8=cltypes.int8
int16=cltypes.int16
long2=cltypes.long2
long4=cltypes.long4
long8=cltypes.long8
long16=cltypes.long16
uchar2=cltypes.uchar2
uchar4=cltypes.uchar4
uchar8=cltypes.uchar8
uchar16=cltypes.uchar16
ushort2=cltypes.ushort2
ushort4=cltypes.ushort4
ushort8=cltypes.ushort8
ushort16=cltypes.ushort16
uint2=cltypes.uint2
uint4=cltypes.uint4
uint8=cltypes.uint8
uint16=cltypes.uint16
ulong2=cltypes.ulong2
ulong4=cltypes.ulong4
ulong8=cltypes.ulong8
ulong16=cltypes.ulong16
half2=cltypes.half2
half4=cltypes.half4
half8=cltypes.half8
half16=cltypes.half16
float2=cltypes.float2
float4=cltypes.float4
float8=cltypes.float8
float16=cltypes.float16
double2=cltypes.double2
double4=cltypes.double4
double8=cltypes.double8
double16=cltypes.double16
char=cltypes.char
short=cltypes.short
int=cltypes.int
long=cltypes.long
uchar=cltypes.uchar
ushort=cltypes.ushort
uint=cltypes.uint
ulong=cltypes.ulong
half=cltypes.half
float=cltypes.float
double=cltypes.double

cfloat=np.dtype(np.complex64)
cdouble=np.dtype(np.complex128)
                         