from pyopencl_extension.modifications_pyopencl import cltypes
from pyopencl_extension.types.type_handler import *
int_ = int
float_=float
char=TypeHandlerScalar(cltypes.char)
short=TypeHandlerScalar(cltypes.short)
int=TypeHandlerScalar(cltypes.int)
long=TypeHandlerScalar(cltypes.long)
uchar=TypeHandlerScalar(cltypes.uchar)
ushort=TypeHandlerScalar(cltypes.ushort)
uint=TypeHandlerScalar(cltypes.uint)
ulong=TypeHandlerScalar(cltypes.ulong)
half=TypeHandlerScalar(cltypes.half)
float=TypeHandlerScalar(cltypes.float)
double=TypeHandlerScalar(cltypes.double)
char2=TypeHandlerVec('char2')
char4=TypeHandlerVec('char4')
char8=TypeHandlerVec('char8')
char16=TypeHandlerVec('char16')
short2=TypeHandlerVec('short2')
short4=TypeHandlerVec('short4')
short8=TypeHandlerVec('short8')
short16=TypeHandlerVec('short16')
int2=TypeHandlerVec('int2')
int4=TypeHandlerVec('int4')
int8=TypeHandlerVec('int8')
int16=TypeHandlerVec('int16')
long2=TypeHandlerVec('long2')
long4=TypeHandlerVec('long4')
long8=TypeHandlerVec('long8')
long16=TypeHandlerVec('long16')
uchar2=TypeHandlerVec('uchar2')
uchar4=TypeHandlerVec('uchar4')
uchar8=TypeHandlerVec('uchar8')
uchar16=TypeHandlerVec('uchar16')
ushort2=TypeHandlerVec('ushort2')
ushort4=TypeHandlerVec('ushort4')
ushort8=TypeHandlerVec('ushort8')
ushort16=TypeHandlerVec('ushort16')
uint2=TypeHandlerVec('uint2')
uint4=TypeHandlerVec('uint4')
uint8=TypeHandlerVec('uint8')
uint16=TypeHandlerVec('uint16')
ulong2=TypeHandlerVec('ulong2')
ulong4=TypeHandlerVec('ulong4')
ulong8=TypeHandlerVec('ulong8')
ulong16=TypeHandlerVec('ulong16')
half2=TypeHandlerVec('half2')
half4=TypeHandlerVec('half4')
half8=TypeHandlerVec('half8')
half16=TypeHandlerVec('half16')
float2=TypeHandlerVec('float2')
float4=TypeHandlerVec('float4')
float8=TypeHandlerVec('float8')
float16=TypeHandlerVec('float16')
double2=TypeHandlerVec('double2')
double4=TypeHandlerVec('double4')
double8=TypeHandlerVec('double8')
double16=TypeHandlerVec('double16')